@@ -1,47 +1,119 @@
 import os
 import subprocess
-import contextlib
 import multiprocessing
-from .atomicwrite import atomic_write
 import tempfile
 import sys
 import random
-import shutil
+import contextlib
 
 from .fasta import fasta_iter
 
+@contextlib.contextmanager
+def possibly_compressed_write(filename):
+    from .atomicwrite import atomic_write
+    if filename.endswith('.gz'):
+        import gzip
+        mode = 'wb'
+        transf = lambda f : gzip.open(f, mode='wt')
+    elif filename.endswith('.bz2'):
+        import bz2
+        mode = 'wb'
+        transf = lambda f : bz2.open(f, mode='wt')
+    elif filename.endswith('.xz'):
+        import lzma
+        mode = 'wb'
+        transf = lambda f : lzma.open(f, mode='wt')
+    else:
+        mode = 'wt'
+        transf = lambda f : f
+    with atomic_write(filename, mode=mode, overwrite=True) as f:
+        g = transf(f)
+        yield g
+        if g is not f:
+            g.close()
+
+def check_training_mode(logger, args):
+    if args.is_semibin2:
+        args.training_type = 'auto'
+
+    if args.training_type == 'semi' and args.self_supervised:
+        logger.error('Both --training-type=semi and --self-supervised arguments used')
+        sys.exit(1)
+
+    if args.training_type == 'self' and args.semi_supervised:
+        logger.warning('Both --training-type=self and --semi-supervised arguments used')
+        sys.exit(1)
+
+    if not args.self_supervised and not args.semi_supervised:
+        if args.training_type == 'self' or args.is_semibin2:
+            logger.info(
+                f"SemiBin will run in self supervised mode")
+            args.training_type = 'self'
+        else:
+            logger.info(
+                f"SemiBin will run in semi supervised mode")
+            args.training_type = 'semi'
+
+    elif args.self_supervised and args.semi_supervised:
+        if args.is_semibin2:
+            logger.warning(
+                f'You chose both semi-supervised and self-supervised learning! SemiBin will use self-supervised learning')
+            args.training_type = 'self'
+        else:
+            logger.warning(
+                f'You chose both semi-supervised and self-supervised learning! SemiBin will use semi-supervised learning (this may change in the future)')
+            args.training_type = 'semi'
+
+    elif args.self_supervised:
+        args.training_type = 'self'
+    else:
+        args.training_type = 'semi'
+
+
 def validate_normalize_args(logger, args):
     '''Validate and normalize command line arguments'''
+    exit_with_error = False
+
     def expect_file(f):
+        nonlocal exit_with_error
         if f is not None:
             if not os.path.exists(f):
                 sys.stderr.write(
                     f"Error: Expected file '{f}' does not exist\n")
-                sys.exit(1)
+                exit_with_error = True
 
     def expect_file_list(fs):
         if fs is not None:
             for f in fs:
                 expect_file(f)
 
-    if args.cmd != 'download_GTDB' and args.cmd != 'check_install' and args.cmd != 'concatenate_fasta':
+    if hasattr(args, 'num_process'):
         if args.num_process == 0:
             args.num_process = multiprocessing.cpu_count()
             logger.info(f'Setting number of CPUs to {args.num_process}')
 
         if args.num_process > multiprocessing.cpu_count():
             args.num_process = multiprocessing.cpu_count()
+        os.environ['NUMEXPR_NUM_THREADS'] = str(args.num_process)
+        os.environ['NUMEXPR_MAX_THREADS'] = str(args.num_process)
+        os.environ['OMP_NUM_THREADS'] = str(args.num_process)
 
-    if args.cmd in ['single_easy_bin', 'multi_easy_bin', 'train', 'bin', 'train_self']:
-        if args.cmd != 'train_self':
-            if args.orf_finder not in ['prodigal', 'fraggenescan']:
-                sys.stderr.write(
-                    f"Error: SemiBin only supports prodigal or fraggenescan as the ORF finder (--orf_finder option).\n")
-                sys.exit(1)
+
+    if args.cmd in ['single_easy_bin', 'multi_easy_bin', 'train', 'bin']:
+        if args.orf_finder not in ['prodigal', 'fraggenescan', 'fast-naive']:
+            sys.stderr.write(
+                f"Error: SemiBin only supports 'prodigal'/'fraggenescan'/'fast-naive' as the ORF finder (--orf_finder option).\n")
+            exit_with_error = True
+        expect_file(args.prodigal_output_faa)
+        if args.prodigal_output_faa is not None:
+            args.orf_finder = 'prodigal'
+
+    if hasattr(args, 'engine'):
+        args.engine = args.engine.lower()
         if args.engine not in ['auto', 'gpu', 'cpu']:
             sys.stderr.write(
-                f"Error: You need to specify the engine in auto/gpu/cpu.\n")
-            sys.exit(1)
+                f"Error: Argument '--engine' needs to be one of auto[default]/gpu/cpu.\n")
+            exit_with_error = True
 
     if args.cmd == 'generate_cannot_links':
         if args.GTDB_reference is not None:
@@ -58,33 +130,33 @@ def validate_normalize_args(logger, args):
         expect_file_list(args.bams)
 
     if args.cmd in ['train', 'train_self']:
-        if args.mode == 'single':
+        if not args.train_from_many:
             if len(args.data) > 1:
                 sys.stderr.write(
                     f"Error: Expected one data.csv file with single mode.\n")
-                sys.exit(1)
+                exit_with_error = True
 
             if len(args.data_split) > 1:
                 sys.stderr.write(
                     f"Error: Expected one data_split.csv file with single mode.\n")
-                sys.exit(1)
+                exit_with_error = True
             if args.cmd == 'train':
                 if len(args.contig_fasta) > 1:
                     sys.stderr.write(
                         f"Error: Expected one fasta file with single mode.\n")
-                    sys.exit(1)
+                    exit_with_error = True
 
                 if len(args.cannot_link) > 1:
                     sys.stderr.write(
                         f"Error: Expected one cannot.txt file with single mode.\n")
-                    sys.exit(1)
+                    exit_with_error = True
                 expect_file(args.cannot_link[0])
                 expect_file(args.contig_fasta[0])
 
             expect_file(args.data[0])
             expect_file(args.data_split[0])
 
-        elif args.mode == 'several':
+        else:
             if args.cmd == 'train':
                 assert len(args.contig_fasta) == len(args.data) == len(args.data_split) == len(args.cannot_link), 'Must input same number of fasta, data, data_split, cannot files!'
                 expect_file_list(args.cannot_link)
@@ -95,21 +167,26 @@ def validate_normalize_args(logger, args):
             expect_file_list(args.data)
             expect_file_list(args.data_split)
 
-
+    if args.cmd in ['single_easy_bin', 'multi_easy_bin']:
+        if args.sequencing_type.lower() in ['short', 'short-read', 'short_reads', 'short-reads', 'short_read']:
+            args.sequencing_type = 'short_read'
+        elif args.sequencing_type.lower() in ['long', 'long-read', 'long_reads', 'long-reads', 'long_read']:
+            args.sequencing_type = 'long_read'
         else:
             sys.stderr.write(
-                f"Error: Please use training mode with [single/several].\n")
-            sys.exit(1)
+                f"Error: Did not understand sequencing_type argument '{args.sequencing_type}' (should be short_reads or long_reads).\n")
+            exit_with_error = True
+        logger.info(f'Binning for {args.sequencing_type}')
 
     if args.cmd == 'bin':
         if args.environment is None and args.model_path is None:
             sys.stderr.write(
                 f"Error: Please choose input a model path or use our built-in model for [human_gut/dog_gut/ocean/soil/cat_gut/human_oral/mouse_gut/pig_gut/built_environment/wastewater/chicken_caecum/global].\n")
-            sys.exit(1)
+            exit_with_error = True
         if args.environment is not None and args.model_path is not None:
             sys.stderr.write(
                 f"Error: Please choose input a model path or use our built-in model for [human_gut/dog_gut/ocean/soil/cat_gut/human_oral/mouse_gut/pig_gut/built_environment/wastewater/chicken_caecum/global].\n")
-            sys.exit(1)
+            exit_with_error = True
         if args.environment is not None:
             # This triggers checking that the environment is valid
             get_model_path(args.environment)
@@ -127,13 +204,11 @@ def validate_normalize_args(logger, args):
         if args.environment is not None:
             # This triggers checking that the environment is valid
             get_model_path(args.environment)
-
-        if args.training_type not in ['semi', 'self']:
-            sys.stderr.write(
-                f"Error: You need to specify the training algorithm in semi/self.\n")
-            sys.exit(1)
+        else:
+            check_training_mode(logger, args)
 
     if args.cmd == 'multi_easy_bin':
+        check_training_mode(logger, args)
         if args.GTDB_reference is not None:
             expect_file(args.GTDB_reference)
         expect_file(args.contig_fasta)
@@ -142,7 +217,7 @@ def validate_normalize_args(logger, args):
         if args.training_type not in ['semi', 'self']:
             sys.stderr.write(
                 f"Error: You need to specify the training algorithm in semi/self.\n")
-            sys.exit(1)
+            exit_with_error = True
 
     if args.cmd == 'concatenate_fasta':
         expect_file_list(args.contig_fasta)
@@ -152,8 +227,13 @@ def validate_normalize_args(logger, args):
         if len(set(contig_name)) != len(contig_name):
             sys.stderr.write(
                 f"Error: Make sure that every contig file have different names.\n")
-            sys.exit(1)
+            exit_with_error = True
 
+    if getattr(args, 'train_from_many', False):
+        args.mode = 'several'
+
+    if exit_with_error:
+        sys.exit(1)
 
 
 def get_must_link_threshold(contig_lens):
@@ -241,7 +321,9 @@ normalize_marker_trans__dict = {
     'TIGR02386': 'TIGR02387',
 }
 
-def get_marker(hmmout, fasta_path=None, min_contig_len=None, multi_mode=False, orf_finder = None):
+def get_marker(hmmout, fasta_path=None, min_contig_len=None, multi_mode=False, orf_finder = None, contig_to_marker = False):
+    '''Parse HMM output file and return markers
+    '''
     import pandas as pd
     data = pd.read_table(hmmout, sep=r'\s+',  comment='#', header=None,
                          usecols=(0,3,5,15,16), names=['orf', 'gene', 'qlen', 'qstart', 'qend'])
@@ -251,7 +333,7 @@ def get_marker(hmmout, fasta_path=None, min_contig_len=None, multi_mode=False, o
     qlen = data[['gene','qlen']].drop_duplicates().set_index('gene')['qlen']
 
     def contig_name(ell):
-        if orf_finder == 'prodigal':
+        if orf_finder in ['prodigal', 'fast-naive']:
             contig,_ = ell.rsplit( '_', 1)
         else:
             contig,_,_,_ = ell.rsplit( '_', 3)
@@ -264,117 +346,42 @@ def get_marker(hmmout, fasta_path=None, min_contig_len=None, multi_mode=False, o
         data = data[data['contig'].map(lambda c: contig_len[c] >= min_contig_len)]
     data = data.drop_duplicates(['gene', 'contig'])
 
-    def extract_seeds(vs, sel):
-        vs = vs.sort_values()
-        median = vs[len(vs) //2]
-
-        # the original version broke ties by picking the shortest query, so we
-        # replicate that here:
-        candidates = vs.index[vs == median]
-        c = qlen.loc[candidates].idxmin()
-        r = list(sel.query('gene == @c')['contig'])
-        r.sort()
-        return r
-
-
-    if multi_mode:
-        data['bin'] = data['orf'].str.split('.', n=0, expand=True)[0]
-        counts = data.groupby(['bin', 'gene'])['orf'].count()
-        res = {}
-        for b,vs in counts.groupby(level=0):
-            cs = extract_seeds(vs.droplevel(0), data.query('bin == @b', local_dict={'b':b}))
-            res[b] = [c.split('.',1)[1] for c in cs]
-        return res
+    if contig_to_marker:
+        from collections import defaultdict
+        marker = data['gene'].values
+        contig = data['contig'].values
+        sequence2markers = defaultdict(list)
+        for m, c in zip(marker, contig):
+            sequence2markers[c].append(m)
+        return sequence2markers
     else:
-        counts = data.groupby('gene')['orf'].count()
-        return extract_seeds(counts, data)
+        def extract_seeds(vs, sel):
+            vs = vs.sort_values()
+            median = vs[len(vs) //2]
 
-def prodigal(contig_file, contig_output):
-        with open(contig_output + '.out', 'w') as prodigal_out_log:
-            subprocess.check_call(
-                ['prodigal',
-                 '-i', contig_file,
-                 '-p', 'meta',
-                 '-q',
-                 '-m', # See https://github.com/BigDataBiology/SemiBin/issues/87
-                 '-a', contig_output
-                 ],
-                stdout=prodigal_out_log,
-            )
+            # the original version broke ties by picking the shortest query, so we
+            # replicate that here:
+            candidates = vs.index[vs == median]
+            c = qlen.loc[candidates].idxmin()
+            r = list(sel.query('gene == @c')['contig'])
+            r.sort()
+            return r
 
-def run_prodigal(fasta_path, num_process, output):
-    from .error import LoggingPool
 
-    contigs = {}
-    for h, seq in fasta_iter(fasta_path):
-        contigs[h] = seq
+        if multi_mode:
+            data['bin'] = data['orf'].str.split(pat='.', n=0, expand=True)[0]
+            counts = data.groupby(['bin', 'gene'])['orf'].count()
+            res = {}
+            for b,vs in counts.groupby(level=0):
+                cs = extract_seeds(vs.droplevel(0), data.query('bin == @b', local_dict={'b':b}))
+                res[b] = [c.split('.',1)[1] for c in cs]
+            return res
+        else:
+            counts = data.groupby('gene')['orf'].count()
+            return extract_seeds(counts, data)
 
-    total_len = sum(len(s) for s in contigs.values())
-    split_len = total_len // num_process
 
-    cur = split_len + 1
-    next_ix = 0
-    out = None
-    with contextlib.ExitStack() as stack:
-        for h,seq in contigs.items():
-            if cur > split_len and next_ix < num_process:
-                if out is not None:
-                    out.close()
-                out = open(os.path.join(output, 'contig_{}.fa'.format(next_ix)), 'wt')
-                out = stack.enter_context(out)
-
-                cur = 0
-                next_ix += 1
-            out.write(f'>{h}\n{seq}\n')
-            cur += len(seq)
-
-    with LoggingPool(num_process) if num_process != 0 else LoggingPool() as pool:
-        try:
-            for index in range(next_ix):
-                pool.apply_async(
-                    prodigal,
-                    args=(
-                        os.path.join(output, 'contig_{}.fa'.format(index)),
-                        os.path.join(output, 'contig_{}.faa'.format(index)),
-                    ))
-            pool.close()
-            pool.join()
-        except:
-            sys.stderr.write(
-                f"Error: Running prodigal fail\n")
-            sys.exit(1)
-
-    contig_output = os.path.join(output, 'contigs.faa')
-    with open(contig_output, 'w') as f:
-        for index in range(next_ix):
-            f.write(open(os.path.join(output, 'contig_{}.faa'.format(index)), 'r').read())
-    return contig_output
-
-def run_fraggenescan(fasta_path, num_process, output):
-    try:
-        contig_output = os.path.join(output, 'contigs.faa')
-        with open(contig_output + '.out', 'w') as frag_out_log:
-            # We need to call FragGeneScan instead of the Perl wrapper because the
-            # Perl wrapper does not handle filepaths correctly if they contain spaces
-            # This binary does not handle return codes correctly, though, so we
-            # cannot use `check_call`:
-            subprocess.call(
-                [shutil.which('FragGeneScan'),
-                 '-s', fasta_path,
-                 '-o', contig_output,
-                 '-w', str(0),
-                 '-t', 'complete',
-                 '-p', str(num_process),
-                 ],
-                stdout=frag_out_log,
-            )
-    except:
-        sys.stderr.write(
-            f"Error: Running fraggenescan failed\n")
-        sys.exit(1)
-    return contig_output + '.faa'
-
-def cal_num_bins(fasta_path, binned_length, num_process, multi_mode=False, output = None, orf_finder = 'prodigal'):
+def cal_num_bins(fasta_path, binned_length, num_process, multi_mode=False, output = None, orf_finder = 'prodigal', prodigal_output_faa=None):
     '''Estimate number of bins from a FASTA file
 
     Parameters
@@ -383,6 +390,7 @@ def cal_num_bins(fasta_path, binned_length, num_process, multi_mode=False, outpu
     num_process: int (number of CPUs to use)
     multi_mode: bool, optional (if True, treat input as resulting from concatenating multiple files)
     '''
+    from .orffinding import run_orffinder
     with tempfile.TemporaryDirectory() as tdir:
         if output is not None:
             if os.path.exists(os.path.join(output, 'markers.hmmout')):
@@ -393,8 +401,7 @@ def cal_num_bins(fasta_path, binned_length, num_process, multi_mode=False, outpu
         else:
             target_dir = tdir
 
-        run_orffinder = run_prodigal if orf_finder == 'prodigal' else run_fraggenescan
-        contig_output = run_orffinder(fasta_path, num_process, tdir)
+        contig_output = run_orffinder(fasta_path, num_process, tdir, orf_finder, prodigal_output_faa=prodigal_output_faa)
 
         hmm_output = os.path.join(target_dir, 'markers.hmmout')
         try:
@@ -417,13 +424,11 @@ def cal_num_bins(fasta_path, binned_length, num_process, multi_mode=False, outpu
                 f"Error: Running hmmsearch fail\n")
             sys.exit(1)
 
-        marker = get_marker(hmm_output, fasta_path, binned_length, multi_mode, orf_finder=orf_finder)
-
-        return marker
+        return get_marker(hmm_output, fasta_path, binned_length, multi_mode, orf_finder=orf_finder)
 
 
 def write_bins(namelist, contig_labels, output, contig_seqs,
-               recluster=False, origin_label=0, minfasta = 200000):
+               minfasta = 200000, output_tag=None, output_compression='none'):
     '''
     Write binned FASTA files
 
@@ -445,12 +450,16 @@ def write_bins(namelist, contig_labels, output, contig_seqs,
         whole_bin_bp = sum(sizes)
 
         if whole_bin_bp >= minfasta:
-            ofname = f'bin.{label}.fa' if not recluster \
-                    else f'recluster_{origin_label}.bin.{label}.fa'
+            if output_tag is None:
+                ofname = f'bin.{label}.fa'
+            else:
+                ofname = f'{output_tag}_{label}.fa'
             ofname = os.path.join(output, ofname)
+            if output_compression != 'none':
+                ofname += '.' + output_compression
             n50, l50 = n50_l50(sizes)
             written.append([ofname, whole_bin_bp, len(contigs), n50, l50])
-            with atomic_write(ofname, overwrite=True) as ofile:
+            with possibly_compressed_write(ofname) as ofile:
                 for contig in contigs:
                     ofile.write(f'>{contig}\n{contig_seqs[contig]}\n')
     return pd.DataFrame(written, columns=['filename', 'nbps', 'nr_contigs', 'N50', 'L50'])
